# Inventory ... this should probably be JSON file

SKLEARNED_CHAMPION_URLS = ['https://raw.githubusercontent.com/microprediction/sklearned/main/champion_models/tsa_aggressive_d0_ensemble_1_160.json']








from momentum.skatertools.parade import parade
from successor.extension.reflection import reflect
from successor.skaters.scalarskaters.sklearnedio import get_local_compiled_model
from successor.interpolation.linear import linear_interpolator
from successor.conventions import wrap, nonecast
import numpy as np
import time



def scaler_skater_factory(y, s, k:int, skater_name:str, n_input:int, extender=None, interpolator=None):
    """ Runs keras surrogate models, extending the time-series as required.

            skater_name
            n_input
            extender    : Takes [ float ] len<n_input possibly, to array (1, 1, n_input)

        This will try to find local models for each k. If it cannot, it will interpolate or extrapolate based on the
        surrogates that it is able to find. By default interpolation is linear and extrapolation is nearest neighbour.

    """
    if extender is None:
        extender = reflect

    if interpolator is None:
        interpolator = linear_interpolator

    y0 = wrap(y)[0]
    if not s.get('p'):
        init_start_time = time.time()
        s = {'p':{},
             'models':dict(),
             'y':[]}
        n_models = 0
        for model_k in range(1,k+1):
            try:
                s['models'][model_k] = get_local_compiled_model(skater_name=skater_name,k=model_k,n_input=n_input)
                n_models += 1
            except:
                print('No surrogate found for k='+str(k)+' so interpolation or extrapolation will be used.')
        if n_models==0:
            raise LookupError('Cannot instantiate as no local model was found for '+skater_name)
        s['cpu'] = {'initialization':time.time()-init_start_time,'invocation':0,'count':0}
    if y0 is None:
        return None, s, None
    else:
        invocation_start_time = time.time()
        s['y'].append(y0)

        # Use neural networks to predict
        y_extended = extender(s['y'][:n_input],n_input=n_input)
        if not len(y_extended)==n_input:
            raise IndexError('wrong length returned by extender')
        y_extended_input = np.ndarray(shape=(1,1,n_input))
        y_extended_input[0,0,:] = y_extended
        x_hat = [np.nan for _ in range(k)]
        for model_k, k_model in s['models'].items():
            x = k_model(y_extended_input) # faster then k_model.predict( )
                                # https://github.com/tensorflow/tensorflow/commit/42f469be0f3e8c36624f0b01c571e7ed15f75faf
            x_hat[model_k-1] = x[0,0,0]

        # Interpolate
        x = interpolator(x_hat)

        # Use empirical x_std, ignore empirical bias
        _x_bias, x_std, s['p'] = parade(p=s['p'], x=x, y=y0)
        x_std_fallback = nonecast(x_std,fill_value=1.0)

        s['cpu']['invocation']+= time.time()-invocation_start_time
        s['cpu']['count']+=1
        return x, x_std_fallback, s










import pathlib
import os
from getjson import getjson
from pprint import pprint
import json
from successor.skaters.scalarskaters.sklearnedinventory import SKLEARNED_CHAMPION_URLS
from tensorflow import keras
from successor.conventions import keras_optimizer_from_name

ROOT_PATH = str(pathlib.Path(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent)
MODEL_CACHE = ROOT_PATH + os.path.sep + 'sklearnedmodelcache'
INFO_CACHE = ROOT_PATH + os.path.sep + 'sklearnedinfocache'
SKL_REMOTE_PATH = 'https://raw.githubusercontent.com/microprediction/sklearned/main'
SKL_REMOTE_MODEL_PATH = SKL_REMOTE_PATH + '/champion_models'
SKL_REMOTE_INFO_PATH = SKL_REMOTE_PATH + '/champion_info'
SKL_EXAMPLE_REMOTE_PATH = SKL_REMOTE_MODEL_PATH + os.path.sep + 'tsa_aggressive_combined_ensemble_1_80.json'


def champion_from_url(url):
    # e.g. https: // raw.githubusercontent.com / microprediction / sklearned / main / champion_models / tsa_aggressive_combined_ensemble_1_80.json
    last_bit = url.split('/')[-1].replace('.json','')
    n_input = int(last_bit.split('_')[-1])
    k = int(last_bit.split('_')[-2])
    skater_name = '_'.join( last_bit.split('_')[:-2])
    return dict(skater_name=skater_name,k=k,n_input=n_input)

SKLEARNED_CHAMPIONS = [champion_from_url(url) for url in SKLEARNED_CHAMPION_URLS]


def champion_json(skater_name,k,n_input):
    return skater_name + '_' + str(k) + '_' + str(n_input) + '.json'


def remote_info_path(skater_name, k: int, n_input: int):
    return SKL_REMOTE_INFO_PATH + '/' + champion_json(skater_name=skater_name,k=k,n_input=n_input)


def remote_model_path(skater_name, k: int, n_input: int):
    return SKL_REMOTE_MODEL_PATH + '/' + champion_json(skater_name=skater_name, k=k, n_input=n_input)


def local_model_path(skater_name, k: int, n_input: int):
    return MODEL_CACHE + os.path.sep + champion_json(skater_name=skater_name,k=k,n_input=n_input)


def local_info_path(skater_name, k: int, n_input: int):
    return INFO_CACHE + os.path.sep + champion_json(skater_name=skater_name,k=k,n_input=n_input)



def _get_remote_model(skater_name:str, k: int, n_input: int):
    _path = remote_model_path(skater_name=skater_name, k=k, n_input=n_input)
    _res = getjson(_path)
    if isinstance(_res,str):
        return json.loads(_res)
    elif isinstance(_res,dict):
        return _res
    else:
        print('Missing '+remote_model_path(skater_name=skater_name,k=k,n_input=n_input))


def _get_remote_info(skater_name:str, k: int, n_input: int):
    _path = remote_info_path(skater_name=skater_name, k=k, n_input=n_input)
    _res = getjson(_path)
    if isinstance(_res,str):
        return json.loads(_res)
    elif isinstance(_res,dict):
        return _res
    else:
        print('Missing '+remote_info_path(skater_name=skater_name,k=k,n_input=n_input))


def get_local_model(skater_name:str, k: int, n_input: int):
    with open(local_model_path(skater_name=skater_name, k=k, n_input=n_input),'rt') as fp:
        model_json = fp.read()
    model = keras.models.model_from_json(model_json)
    return model

def get_local_info(skater_name:str, k: int, n_input: int):
    with open(local_info_path(skater_name=skater_name, k=k, n_input=n_input),'rt') as fp:
        model = json.load(fp)
    return model


def save_model_locally(model: dict, skater_name: str, k: int, n_input: int):
    _file = local_model_path(skater_name=skater_name, k=k, n_input=n_input)
    with open(_file, 'wt') as fp:
        json.dump(obj=model,fp=fp)


def save_info_locally(info: dict, skater_name: str, k: int, n_input: int):
    _file = local_info_path(skater_name=skater_name, k=k, n_input=n_input)
    with open(_file, 'wt') as fp:
        json.dump(obj=info,fp=fp)


def copy_inventory_to_cache():
    for champ in SKLEARNED_CHAMPIONS:
        model = _get_remote_model(**champ)
        info  = _get_remote_info(**champ)
        if model is not None and info is not None:
            save_model_locally(model=model,**champ)
            save_info_locally(info=info, **champ)


def get_local_compiled_model(skater_name:str,k:int,n_input:int):
    model = get_local_model(skater_name=skater_name,k=k,n_input=n_input)
    model_info = get_local_info(skater_name=skater_name,k=k,n_input=n_input)
    keras_optimizer = keras_optimizer_from_name(opt_name=model_info['keras_optimizer'], learning_rate=model_info['learning_rate'])
    model.compile(optimizer=keras_optimizer)
    return model


if __name__ == '__main__':
    print('example model location is ' + SKL_EXAMPLE_REMOTE_PATH)
    copy_inventory_to_cache()
    n_input = 160
    model = get_local_compiled_model(skater_name='tsa_aggressive_d0_ensemble',k=1,n_input=n_input)
    pprint(model)
    import numpy as np
    x = np.random.randn(300,1,n_input)
    y = model.predict(x)
    pass



from successor.skaters.scalarskaters.scalarskaterfactory import scaler_skater_factory


def successor_tsa_aggressive_d0_ensemble(y,s,k,a=None,t=None,e=None,r=None):
    return scaler_skater_factory(y=y,s=s,k=k,skater_name='tsa_aggressive_d0_ensemble',n_input=160)


SCALAR_TSA_SKATERS = [successor_tsa_aggressive_d0_ensemble]



# Borrowed from timemachines

def nonecast(x,fill_value=0.0):
    if x is not None:
        return [xj if xj is not None else fill_value for xj in x]

def wrap(x):
    """ Ensure x is a list of float """
    if x is None:
        return None
    elif isinstance(x,(float,int)):
        return [float(x)]
    else:
        return list(x)


from tensorflow import keras

# TODO: Move into package shared by sklearned, maybe


def keras_optimizer_from_name(opt_name,learning_rate):
    if opt_name == 'SGD':
        return keras.optimizers.SGD(learning_rate=learning_rate)
    elif opt_name == 'RMSprop':
        return keras.optimizers.RMSprop(learning_rate=learning_rate)
    elif opt_name == 'Adam':
        return keras.optimizers.Adam(learning_rate=learning_rate)
    elif opt_name == 'Adagrad':
        return keras.optimizers.Adagrad(learning_rate=learning_rate)
    elif opt_name == 'Adadelta':
        return keras.optimizers.Adagrad(learning_rate=learning_rate)
    elif opt_name == 'Adamax':
        return  keras.optimizers.Adamax(learning_rate=learning_rate)
    elif opt_name == 'Nadam':
        return keras.optimizers.Nadam(learning_rate=learning_rate)
    elif opt_name =='Ftrl':
        return keras.optimizers.Ftrl(learning_rate=learning_rate)
    else:
        raise ValueError('Forgot '+opt_name)
